package lpll;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Enumeration;

import weka.core.Attribute;
import weka.core.DenseInstance;
import weka.core.Instance;
import weka.core.Instances;
import weka.core.Utils;
import weka.core.converters.ConverterUtils.DataSource;


public class lplist {
	//parametrization for learning the LP list, set 3 as default
	int maxGrouping = 3;
	//monotony indicator
    static int [] monotony; 
    //the LP list represented by the attributes, preferences on the attribute values and the respective cut points for the attribute values
	//the preferences on attribute(combination) 
	int[][] attributes;
	//the LP-Preference "List" including the local preference rules and 
	int[][][] preferences;
	//the cutting points for discretization
	double[][][]  cutPoints;
	//validation measures
	int[][] decidedAtNode; 	//array contains correctly [][0] and wrongly [][1] decided instances at each stage
	int errorsGlobal = -1;
	int abstentionsGlobal = 0;
	
	
    
    /**
	 * Constructor
	 */
	public lplist (Instances data, int maxGroup,int bins, int[] monotone) throws Exception {
		//the maximal number of attributes that can be grouped
		maxGrouping = maxGroup;
		//initialize monotony indicator
		monotony = new int[monotone.length];
		for(int i = 0; i < monotony.length; i++){
			monotony[i] = monotone[i];
		}
		
		//Initialize the attribute, cutting points and preference arrays
		attributes = new int[data.numAttributes()/2][];
		cutPoints = new double[data.numAttributes()/2][][];
	   	preferences = new int[data.numAttributes()/2][][];
		
	   	//if grouping should be considered refer to lookahead procedure for attribute grouping
		if(maxGroup > 1){
			Instances lhData = new Instances(data);
			double[] errors = new double[data.numAttributes()/2];
			createLPS_lookahead(data, lhData, 0, bins, 0, errors);
		}
		else{
			createLPStructure(data, 0, bins);
		}
		
	}
    
    
	
	
	/**
	 * main with simple example for learning a LP list (r = 4, g = 3)
	 * @param args
	 */
	public static void main(String[] args) throws Exception {
		//Example for learning a lplist
	    String path = "C:\\01_blood_transfusion";	// USER INPUT;  
	    // Read all the instances in the file (ARFF) for the training data (pairwise preferences required, attribute values non-negative without missing values)
        DataSource source = new DataSource(path +"\\train.arff");	//USER INPUT; training data
	    Instances training = source.getDataSet();
	  	 
	    monotony = new int[training.numAttributes()]; 
		for(int i = 0; i < training.numAttributes()-1; i++){
			 //set monotony indicator 0
			 monotony[i] = 0;	//USER INPUT; include background knowledge about preferences on attribute values if availalbe
		}
		int bins = 3;	// USER INPUT; maximal number of bins (parameter r), default set to 3 (which corresponds to 4 bins
	    lplist exList = new lplist(training, 3, bins, monotony);
	    	 
	     System.out.println("Errors: " +exList.computeErrorsExt(training)[0]);
	     System.out.println("Abstentions: " +exList.computeErrorsExt(training)[1]);
		
  
	}
    
    
	 /**
    * Greedy algorithm for identifying and creating the lp-structure based on the given data
    *@param data The data for evaluation (set of pairwise preferences)
    *@param stage The stage where the node to be labeled is situated
    *@param bins The maximum amount of bins for disrectization
    */
    public void createLPStructure(Instances data, int stage, int bins) throws Exception{
    	int[] bestTmpCandidate = new int[maxGrouping];
    	//the set of possible attribute combination, all possible candidates for the nodes
    	int[][] candidateSet;
    	candidateSet = combinations(maxGrouping, data.numAttributes()/2);
    	//internal measures for errors
    	errorsGlobal = -1;
    	abstentionsGlobal = 0;
    	//Choose attribute for node n s.t. consistent with examples, maximizing correctness
    	for(int i = 0 ; i < candidateSet.length; i++){
    		//create the candidate combination
    		attributes[stage] = new int[candidateSet[i].length];
    		for(int m = 0; m < candidateSet[i].length; m++){
    			attributes[stage][m] = candidateSet[i][m]-1;
    		}
    		//is attribute combination i already in the branch (partly) contained?
		    boolean inBranch = false;
	    	for(int m = 0; m < attributes[stage].length; m++){
	    		if(inBranch(attributes[stage][m], stage)){
	    			inBranch = true;
	    		}
	    	}
	    	if(inBranch == false){
	    		//candidate to be evaluated --> compute local preference rule
	    		Instances tmpData = new Instances(data);
	    		//If there are numeric attribute values --> discretize the attribute's values
	    		cutPoints[stage] = new double[attributes[stage].length][]; 
	    	   	for(int c = 0; c < attributes[stage].length; c++){
	    			if(tmpData.numDistinctValues(attributes[stage][c]) > 0 ) {
	    				//array contains the different attribute values
	    				double tmpArray[] = distAttrDoubleVals(tmpData, attributes[stage][c]);
	    				if( !tmpData.attribute(attributes[stage][c]).isNominal() && tmpArray.length > bins+1){  
	    					cutPoints[stage][c] = cutPoints(attributes[stage][c], tmpData, true, bins);
	    					if(cutPoints[stage][c].length > 0){
	    						tmpData = discretize(attributes[stage][c], cutPoints[stage][c], tmpData);
	    					}
	    					else{
	    						//no boundary points left
	    						cutPoints[stage][c] = cutPoints(attributes[stage][c], tmpData, false, bins);
	    						tmpData = discretize(attributes[stage][c], cutPoints[stage][c], tmpData);
	    					}
	    				}
	    				else if(!tmpData.attribute(attributes[stage][c]).isNominal() && tmpArray.length <= bins ){
	    					//just reclassify attribute values 
    						cutPoints[stage][c] = new double[tmpArray.length+1];
    						for(int l = 0; l < tmpArray.length; l++){
    							cutPoints[stage][c][l] = tmpArray[l];
    						}
    						cutPoints[stage][c][cutPoints[stage][c].length-1] = 0.0;
    						Arrays.sort(cutPoints[stage][c]);
    						tmpData = discretize(attributes[stage][c], cutPoints[stage][c], tmpData);
	    				}
	    				else{
	    					//just reclassify attribute values 
	    					cutPoints[stage][c] = new double[tmpArray.length+1];
	    					for(int l = 0; l < tmpArray.length; l++){
	    						cutPoints[stage][c][l] = tmpArray[l];
	    					}
	    					cutPoints[stage][c][cutPoints[stage][c].length-1] = 0.0;
	    					Arrays.sort(cutPoints[stage][c]);
	    				}
	    			}
	    		}
    			preferences[stage] = orderedAttrVals(tmpData, attributes[stage]);
    			//determine the number of mistakes and abstentions induced by the chosen attribute and preference rule
    			int[] tmpEr = computeErrors(tmpData, stage);
	    		if(errorsGlobal < 0 || tmpEr[0] < errorsGlobal || (tmpEr[0] == errorsGlobal && tmpEr[1] < abstentionsGlobal)){
	    			//current attribute combination and corresponding list is an improvement!
	    			errorsGlobal = tmpEr[0];
	    			abstentionsGlobal = tmpEr[1];
	    			bestTmpCandidate = new int[attributes[stage].length];
	    			for(int node = 0; node <bestTmpCandidate.length; node++){
	    				bestTmpCandidate[node] = attributes[stage][node];
	    			}
	    		}			
	    	}	
	    	//in case that all attributes have already been used)-> create dummy nodes for the path
	    	if(i == candidateSet.length-1 && errorsGlobal < 0){
	    		bestTmpCandidate = new int[1]; 
	    		bestTmpCandidate[0] = 99;
	    	}
	    }
    	//The best candidate attribute to label the node [stage]
    	attributes[stage] = new int[bestTmpCandidate.length];
    	for(int j = 0; j < bestTmpCandidate.length; j++){
    		attributes[stage][j] = bestTmpCandidate[j];
    	}
    	//discretize the candidate's values in the data set
    	cutPoints[stage] = new double[attributes[stage].length][]; 
    	for(int c = 0; c < attributes[stage].length; c++){
    		if(data.numDistinctValues(attributes[stage][c]) > 0 ) {
    			double tmpArray[] = distAttrDoubleVals(data, attributes[stage][c]);
    			if( !data.attribute(attributes[stage][c]).isNominal() && tmpArray.length > bins+1 ){  
    				cutPoints[stage][c] = cutPoints(attributes[stage][c], data, true, bins);
    				data = discretize(attributes[stage][c], cutPoints[stage][c], data);
				}
    			else if(!data.attribute(attributes[stage][c]).isNominal() && tmpArray.length <= bins ){
					//just reclassifiy attribute values 
					cutPoints[stage][c] = new double[tmpArray.length+1];
					for(int l = 0; l < tmpArray.length; l++){
						cutPoints[stage][c][l] = tmpArray[l];
					}
					cutPoints[stage][c][cutPoints[stage][c].length-1] = 0.0;
					Arrays.sort(cutPoints[stage][c]);
					data = discretize(attributes[stage][c], cutPoints[stage][c], data);
				}
    			else{
    				//just reclassify attribute values
    				cutPoints[stage][c] = new double[tmpArray.length+1];
					for(int l = 0; l < tmpArray.length; l++){
						cutPoints[stage][c][l] = tmpArray[l];
					}
					cutPoints[stage][c][cutPoints[stage][c].length-1] = 0.0;
					Arrays.sort(cutPoints[stage][c]);
				}
			}
		}
		if(attributes[stage][0] != 99){
			preferences[stage] = orderedAttrVals(data, attributes[stage]);
	    }
	    //there was no attribute left for consideration, therefore dummy attribute was created 
	    else{
	    	preferences[stage] = new int[1][1];
	    	preferences[stage][0][0] = 99;
	    }
	    //resize the arrays
	    if(attributes[stage].length > 1){
	    	int size = attributes.length - (attributes[stage].length -1);
	    	resize(size);
	    }
	    if(stage+1 < attributes.length){
    		//select the data discarding those instances which are already decided and recursive call 
    		Instance one = new DenseInstance(data.numAttributes()/2);
 			Instance two = new DenseInstance(data.numAttributes()/2);
    		for(int t = 0 ; t < data.numInstances(); t ++){
				for(int j =0; j < data.numAttributes()/2; j++){
					one.setValue(j, data.instance(t).value(j));
					two.setValue(j, data.instance(t).value(j+data.numAttributes()/2));
				}
				if( decide(one, two, stage) != 0 ){
					data.delete(t);
					t = t-1;
				}
    		}
	    	createLPStructure(data, stage+1, bins);
	    }
	}
    
    
    
    
    /**
     * Greedy algorithm for identifying and creating the lp-structure based on the given data
     * Lookahead feature to identify attributes to be grouped together
     *@param data The data for evaluation (set of pairwise preferences)
     *@param lhData The original data stored till lookahead (set of pairwise preferences)
     *@param stage The stage where the node to be labeled is located
     *@param bins The maximum amount of bins for disrectization
     *@param count The number of nodes considered for grouping 
     *@param sumErrors The number of Errors in sum 
     */
    public void createLPS_lookahead(Instances data, Instances lhData, int stage, int bins, int count, double[] sumErrors) throws Exception{
    	count++;
    	int[][] candidateSet;
    	int[] bestTmpCandidate;
    	int[][] tempPreferences;
    	errorsGlobal = -1;
    	abstentionsGlobal = 0;
    	//standard case, no grouping to be considered, yet
    	if(count <= maxGrouping){
    		bestTmpCandidate = new int[1];
    		//the set of possible attributes for the node to be considered
    		candidateSet = combinations(1, data.numAttributes()/2);
    		//choose attribute for node n s.t. consistent with examples, maximizing correctness
    		for(int i = 0 ; i < candidateSet.length; i++){
    			//create the candidate combination
    			attributes[stage] = new int[candidateSet[i].length];
    			for(int m = 0; m < candidateSet[i].length; m++){
    				attributes[stage][m] = candidateSet[i][m]-1;
		    	}
    			//is attribute combination i already in the branch (partly) contained?
    			boolean inBranch = false;
    			for(int m = 0; m < attributes[stage].length; m++){
    				if(inBranch(attributes[stage][m], stage)){
    					inBranch = true;
		    		}
		    	}
    			if(inBranch == false){
    				Instances tmpData = new Instances(data);
    				cutPoints[stage] = new double[attributes[stage].length][]; 
    				for(int c = 0; c < attributes[stage].length; c++){
    					if(tmpData.numDistinctValues(attributes[stage][c]) > 0 ) {
    						double tmpArray[] = distAttrDoubleVals(tmpData, attributes[stage][c]);
    						if( !data.attribute(attributes[stage][c]).isNominal()&& tmpArray.length > bins+1 ){  
    							cutPoints[stage][c] = cutPoints(attributes[stage][c], tmpData, true, bins);
    							if(cutPoints[stage][c].length > 0){
    								tmpData = discretize(attributes[stage][c], cutPoints[stage][c], tmpData);
		    					}
    							else{
    								//no boundary points left
    								cutPoints[stage][c] = cutPoints(attributes[stage][c], tmpData, false, bins);
    								tmpData = discretize(attributes[stage][c], cutPoints[stage][c], tmpData);
    							}
		    				}
    						else if(!data.attribute(attributes[stage][c]).isNominal() && tmpArray.length <= bins+1){ 
    							//just reclassify attribute values 
    							cutPoints[stage][c] = new double[tmpArray.length+1];
    							//copy values identified in the training data set and add 0.0
    							for(int l = 0; l < tmpArray.length; l++){
    								cutPoints[stage][c][l] = tmpArray[l];
		    					}
    							cutPoints[stage][c][cutPoints[stage][c].length-1] = 0.0;
    							Arrays.sort(cutPoints[stage][c]);
    							tmpData = discretize(attributes[stage][c], cutPoints[stage][c], tmpData);
		    				}
    						else {
    							//just reclassify 
    							cutPoints[stage][c] = new double[tmpArray.length+1];
    							for(int l = 0; l < tmpArray.length; l++){
    								cutPoints[stage][c][l] = tmpArray[l];
		    					}
    							cutPoints[stage][c][cutPoints[stage][c].length-1] = 0.0;
    							Arrays.sort(cutPoints[stage][c]);
		    				}
		    			}
		    		}
    				preferences[stage] = orderedAttrVals(tmpData, attributes[stage]);
    				//determine the number of mistakes and abstentions induced by the chosen attribute and preference rule
    				int[] tmpEr = computeErrors(tmpData, stage);
    				if(errorsGlobal < 0 || tmpEr[0] < errorsGlobal || (tmpEr[0] == errorsGlobal && tmpEr[1] < abstentionsGlobal)){
    					//current attribute combination and corresponding tree is an improvement
    					errorsGlobal = tmpEr[0];
    					abstentionsGlobal = tmpEr[1];
    					bestTmpCandidate = new int[attributes[stage].length];
    					for(int node = 0; node <bestTmpCandidate.length; node++){
    						bestTmpCandidate[node] = attributes[stage][node];
		    			}
		    		}		
		    	}
    			//in case that all attributes are already used create dummy nodes for the path
    			if(i == candidateSet.length-1 && errorsGlobal < 0){
		    			//System.out.println("All attributes were already used, dummy node was created!");
		    			bestTmpCandidate = new int[1]; 
		    			bestTmpCandidate[0] = 99;
		    		}
		    	}
		    	//The best attribute to label the node [stage] with is candidate, build the attributes impoartance list
		    	attributes[stage] = new int[bestTmpCandidate.length];
		    	for(int j = 0; j < bestTmpCandidate.length; j++){
		    		attributes[stage][j] = bestTmpCandidate[j];
		    		sumErrors[stage] = errorsGlobal;
		    	}
		    	//discretize the candidate's values in the data set
				cutPoints[stage] = new double[attributes[stage].length][]; 
				for(int c = 0; c < attributes[stage].length; c++){
					if(data.numDistinctValues(attributes[stage][c]) > 0 ) {
						double tmpArray[] = distAttrDoubleVals(data, attributes[stage][c]);
						if( !data.attribute(attributes[stage][c]).isNominal()  && data.numInstances() > bins && tmpArray.length > 2){ 
							cutPoints[stage][c] = cutPoints(attributes[stage][c], data, true, bins);
							data = discretize(attributes[stage][c], cutPoints[stage][c], data);
						}
						else if( !data.attribute(attributes[stage][c]).isNominal() && tmpArray.length <= 2 ){
							//just reclassify attribute values 
							cutPoints[stage][c] = new double[tmpArray.length+1];
							for(int l = 0; l < tmpArray.length; l++){
								cutPoints[stage][c][l] = tmpArray[l];
							}
							cutPoints[stage][c][cutPoints[stage][c].length-1] = 0.0;
							Arrays.sort(cutPoints[stage][c]);
							data = discretize(attributes[stage][c], cutPoints[stage][c], data);
						}
						else{
							//just reclassify attribute values
							cutPoints[stage][c] = new double[tmpArray.length+1];
							for(int l = 0; l < tmpArray.length; l++){
								cutPoints[stage][c][l] = tmpArray[l];
							}
							cutPoints[stage][c][cutPoints[stage][c].length-1] = 0.0;
							Arrays.sort(cutPoints[stage][c]);
						}
					}
				}
		    	if(attributes[stage][0] != 99){
		    		preferences[stage] = orderedAttrVals(data, attributes[stage]);
		    	}
		    	//there was no attribute left for consideration, therefore dummy attribute was created
		    	else{
		    		preferences[stage] = new int[1][1];
		    		preferences[stage][0][0] = 99;
		    	}
		    	//resize the arrays
		    	if(attributes[stage].length > 1){
		    		int size = attributes.length - (attributes[stage].length -1);
		    		resize(size);
		    	}
		    	if(stage+1 < attributes.length){
		    		//select the data w.r.t. to the values which are already decided and recursive call 
		    		Instance one = new DenseInstance(data.numAttributes()/2);
		    		Instance two = new DenseInstance(data.numAttributes()/2);
		    		for(int t = 0 ; t < data.numInstances(); t ++){
		 				for(int j =0; j < data.numAttributes()/2; j++){
		 					one.setValue(j, data.instance(t).value(j));
		 					two.setValue(j, data.instance(t).value(j+data.numAttributes()/2));
			    			}
			    			//Instance can be removed for the next stage
		 				if( decide(one, two, stage) != 0 ){
			    				data.delete(t);
			    				t = t-1;
			    			}
		    		}
		    		createLPS_lookahead(data, lhData, stage+1, bins, count, sumErrors);
		    	}
	    	}
	     	// count = maxGrouping triggers a lookahead for grouping and count/errors need to be reseted 
	    	else{
	    		//jump back for lookahead
	    		stage = stage - maxGrouping;
	    		boolean lhImprove = false;
	    		bestTmpCandidate = new int[maxGrouping];
	    		//the set of possible attribute combination, all possible candidates for the nodes
	    		candidateSet = combinations(maxGrouping, lhData.numAttributes()/2);
	    		//create temporary node to compare error rates
	    		int[] tempAttributes;
	    		double[][] tempCutPoints;
	    		//Choose attribute for node stage s.t. consistent with examples, maximizing correctness
		    	for(int i = 0 ; i < candidateSet.length; i++){
		    		boolean wrongAttrCombo = false;
		    		tempAttributes = new int[candidateSet[i].length];
		    		for(int m = 0; m < candidateSet[i].length; m++){
		    			tempAttributes[m] = candidateSet[i][m]-1;
		    		}
		    		//check requirements w.r.t. attribute combination to be considered
		    		//potential combinations not considering grouping are neglected:
		    		if(candidateSet[i].length < maxGrouping){
		    			wrongAttrCombo = true;
		    		}
		    		//check whether the combination considered contains the lookahead attributes (i.e. attributes not already used before and used lookahead) 
		    		for(int m = 0; m < tempAttributes.length; m++){
		    			//attributes already used -> neglected
		    			if(inBranch(tempAttributes[m], stage)){
		    				wrongAttrCombo = true;
		    			}
		    			//attribute not within the maxGropuing following nodes therefore no lookahead candidate for grouping
		    			if(!(inBranch(tempAttributes[m], stage)) && !(inBranch(tempAttributes[m], (stage+maxGrouping)))){
		    				wrongAttrCombo = true;
		    			}
		    			//ensure also that only feasible attribute combinations are considered w.r.t. monotony, e.g. combinations with two or more monotone attributes are currently not feasible
			    		int ct = 0;
			    		for(int g = 0; g < tempAttributes.length; g++){
			    			if(monotony[tempAttributes[g]]>0){
			    				ct++;
			    			}
			    		}
			    		if(ct>1){
				    		wrongAttrCombo = true;
			    		}
		    		}
		    		if(wrongAttrCombo == false){
		    			//candidate to be evaluated
		    	    	Instances tmpData = new Instances(lhData);
		    	    	tempCutPoints = new double[tempAttributes.length][]; 
		    	    	for(int c = 0; c < tempAttributes.length; c++){
		    				if(tmpData.numDistinctValues(tempAttributes[c]) > 0 ) {
		    					//array contains the different attribute values
		    					double tmpArray[] = distAttrDoubleVals(tmpData, tempAttributes[c]);
		    					if( !data.attribute(tempAttributes[c]).isNominal() && tmpArray.length > bins+1){  
		    						tempCutPoints[c] = cutPoints(tempAttributes[c], tmpData, true, bins);
		    						if(tempCutPoints[c].length > 0){
		    							tmpData = discretize(tempAttributes[c], tempCutPoints[c], tmpData);
		    						}
		    						else{
		    							//no boundary points left
		    							tempCutPoints[c] = cutPoints(tempAttributes[c], tmpData, false, bins);
		    							tmpData = discretize(tempAttributes[c], tempCutPoints[c], tmpData);
		    						}
		    					}
		    					else if( !data.attribute(tempAttributes[c]).isNominal() && tmpArray.length <= bins+1){
		    						//just reclassify attribute values 
		    						tempCutPoints[c] = new double[tmpArray.length+1];
		    						for(int l = 0; l < tmpArray.length; l++){
		    							tempCutPoints[c][l] = tmpArray[l];
		    						}
		    						tempCutPoints[c][tempCutPoints[c].length-1] = 0.0;
		    						Arrays.sort(tempCutPoints[c]);
		    						tmpData = discretize(tempAttributes[c], tempCutPoints[c], tmpData);
		    					}
		    					
		    					
		    					else{
		    						//just reclassify attribute values
		    						tempCutPoints[c] = new double[tmpArray.length+1];
		    						for(int l = 0; l < tmpArray.length; l++){
		    							tempCutPoints[c][l] = tmpArray[l];
		    						}
		    						tempCutPoints[c][tempCutPoints[c].length-1] = 0.0;
		    						Arrays.sort(tempCutPoints[c]);
		    					}
		    				}
		    			}
		    			tempPreferences = orderedAttrVals(tmpData, tempAttributes);
		    			//determine the number of mistakes and abstentions induced by the chosen attribute and preference rule
		    			int[] tmpEr = computeNodeErrors(tmpData, tempAttributes, tempPreferences);
		    			double sumEr = 0;
		    			for(int k = 0; k < maxGrouping; k++){
		    				sumEr = sumEr + sumErrors[stage+k];
		    			}
		    			if(tmpEr[0] < sumEr  && (errorsGlobal < 0 || tmpEr[0] < errorsGlobal || (tmpEr[0] == errorsGlobal && tmpEr[1] < abstentionsGlobal))){
		    				//current attribute combination and corresponding tree is an improvement
		    				errorsGlobal = tmpEr[0];
		    				abstentionsGlobal = tmpEr[1];
		    				bestTmpCandidate = new int[tempAttributes.length];
		    				for(int node = 0; node <bestTmpCandidate.length; node++){
		    					bestTmpCandidate[node] = tempAttributes[node];
		    				}
		    				lhImprove = true;
		    			}		
		    		}
		    	}
		    	//grouping consideration was no improvement --> return to regular processing
		    	if(lhImprove == false){
		    		//select the instances for lhData w.r.t. to the values which are already decided at the former node
		    		if(lhData.attribute(attributes[stage][0]).isNumeric()){
		    			lhData = discretize(attributes[stage][0], cutPoints[stage][0], lhData);	
		    		}
		    		Instance one = new DenseInstance(lhData.numAttributes()/2);
		    		Instance two = new DenseInstance(lhData.numAttributes()/2);
		    		for(int t = 0 ; t < lhData.numInstances(); t ++){
		    			for(int j =0; j < lhData.numAttributes()/2; j++){
		    				one.setValue(j, lhData.instance(t).value(j));
		    				two.setValue(j, lhData.instance(t).value(j+lhData.numAttributes()/2));
		    			}
		    			//Instance can be removed for the next stage
		    			if( decide(one, two, stage) != 0 ){
		    					lhData.delete(t);
		    					t = t-1;
		    			}
		    		}
		    		//reset of parameters 
		    		count = count -2;
		    		stage = stage + maxGrouping;
		    		createLPS_lookahead(data, lhData, stage, bins, count, sumErrors);
		    	}
		    	else{
		    		//The best attribute to label the node [stage] with is candidate, build the attributes impoartance list
			    	attributes[stage] = new int[bestTmpCandidate.length];
			    	for(int j = 0; j < bestTmpCandidate.length; j++){
			    		attributes[stage][j] = bestTmpCandidate[j];
			    	}
			    	//discretize the candidate's values in the data set
					cutPoints[stage] = new double[attributes[stage].length][]; 
	    	    	for(int c = 0; c < attributes[stage].length; c++){
	    				if(lhData.numDistinctValues(attributes[stage][c]) > 0 ) {
	    					//array contains the different attribute values
	    					double tmpArray[] = distAttrDoubleVals(lhData, attributes[stage][c]);
	    					if( !lhData.attribute(attributes[stage][c]).isNominal() && tmpArray.length > 2 && lhData.numInstances() > bins){  
	    						cutPoints[stage][c] = cutPoints(attributes[stage][c], lhData, true, bins);
								lhData = discretize(attributes[stage][c], cutPoints[stage][c], lhData);
	    					}
	    					else if(!lhData.attribute(attributes[stage][c]).isNominal() && tmpArray.length <= 2){ 
	    						//just reclassify attribute values 
								cutPoints[stage][c] = new double[tmpArray.length+1];
								for(int l = 0; l < tmpArray.length; l++){
									cutPoints[stage][c][l] = tmpArray[l];
								}
								cutPoints[stage][c][cutPoints[stage][c].length-1] = 0.0;
								Arrays.sort(cutPoints[stage][c]);
								lhData = discretize(attributes[stage][c], cutPoints[stage][c], lhData);
	    						
	    					}
	    					else{
	    						//just reclassify attribute values 
	    						cutPoints[stage][c] = new double[tmpArray.length+1];
	    						for(int l = 0; l < tmpArray.length; l++){
	    							cutPoints[stage][c][l] = tmpArray[l];
	    						}
	    						cutPoints[stage][c][cutPoints[stage][c].length-1] = 0.0;
	    						Arrays.sort(cutPoints[stage][c]);
	    					}
	    				}
	    			}
					if(attributes[stage][0] != 99){
			    		preferences[stage] = orderedAttrVals(lhData, attributes[stage]);
			    	}
			    	//there was no attribute left for consideration, therefore dummy attribute was created
			    	else{
			    		preferences[stage] = new int[1][1];
			    		preferences[stage][0][0] = 99;
			    	}
			    	//resize the arrays 
			    	if(attributes[stage].length > 1){
			    		int size = attributes.length - (attributes[stage].length -1);
			    		resize(size);
			    	}
			    	//select the data w.r.t. to the values which are already decided and recursive call 
			    	Instance one = new DenseInstance(lhData.numAttributes()/2);
			    	Instance two = new DenseInstance(lhData.numAttributes()/2);
			    	for(int t = 0 ; t < lhData.numInstances(); t ++){
			    		for(int j =0; j < lhData.numAttributes()/2; j++){
			    			one.setValue(j, lhData.instance(t).value(j));
		 					two.setValue(j, lhData.instance(t).value(j+lhData.numAttributes()/2));
			    		}
			    		//Instance can be removed for the next stage
		 				if( decide(one, two, stage) != 0 ){
			    			lhData.delete(t);
			    			t = t-1;
			    		}
			    	}
			    	data = new Instances (lhData);
			    	//reset of parameters 
		    		count = 0;
		    		createLPS_lookahead(data, lhData, stage+1, bins, count, sumErrors);
		    	}
		    }	
	}
    
    
    
    
    /**
	* Returns the number of wrong predictions by applying the rules of the given stage on the data set provided 
	*@param data The data for evaluation
	*@param stage The level of the attribute hierarchy based on which shall be decided
	*@return The number of mistakes [0] induced by the given preference Rules and the number of abstentions [1]
	*/
	public int[] computeErrors(Instances data, int stage){
		boolean same = true;
		int errors = 0;
		int abstentions = 0;
		Instance one = new DenseInstance(data.numAttributes()/2);
		Instance two = new DenseInstance(data.numAttributes()/2);
		//pairwise comparison of the examples
		for(int i = 0; i < data.numInstances(); i++){
			for(int j =0; j < data.numAttributes()/2; j++){
				one.setValue(j, data.instance(i).value(j));
				two.setValue(j, data.instance(i).value(j+data.numAttributes()/2));
				if(one.value(j) != two.value(j)){
					same = false;
				}
			}
			//only compare outcomes that differ
			if(same != true){
				//Instance one is not preferred according to the given rule tree, therefore wrong prediction
				if(decide(one, two, stage) == 2){
					errors = errors +1;
					//abort since no improvement
					if(errorsGlobal > 0 && errorsGlobal < errors){
						int[] result = new int[2];
						result[0] = errors;
						result[1] = abstentions;
						return result;
       				}
				}
				else if (decide(one, two, stage) == 0){
					abstentions = abstentions +1;
				}
			}
   		same = true;
		}
		int[] result = new int[2];
		result[0] = errors;
		result[1] = abstentions;
		return result;
	}
    
	
	
	/**
     * Returns the number of wrong predictions by applying the entire lp-list on the given dataset
     *@param data The data for evaluation
     *@return The number of mistakes [0] induced by the given preference Rules and the number of abstentions [1]
     */
    public int[] computeErrorsExt(Instances data) throws Exception{
    	boolean same = true;
    	int errors = 0;
    	int abstentions = 0;
    	int result;
    	//initialize the array for analyzing nodes
    	decidedAtNode = new int[attributes.length][2];
    	for(int i = 0; i < decidedAtNode.length; i++){
    		decidedAtNode[i][0] =0;
    		decidedAtNode[i][1] =0;
    	}
    	Instance one = new DenseInstance(data.numAttributes()/2);
    	Instance two = new DenseInstance(data.numAttributes()/2);
    	//pairwise comparison of the examples
    	for(int i = 0; i < data.numInstances(); i++){
    		for(int j =0; j < data.numAttributes()/2; j++){
    			one.setValue(j, data.instance(i).value(j));
    			two.setValue(j, data.instance(i).value(j+data.numAttributes()/2));
    			if(one.value(j) != two.value(j)){
    				same = false;
    			}
    		}
    		//compare only outcomes that are not the same
    		if(same == false){
    			result = decideExt(one, two, data);
    			//Instance one is not preferred according to the given rule tree therefore wrong prediction
        		if(result == 2){
        			errors = errors +1;
        		}
        		else if (result == 0){
        			abstentions = abstentions +1;
        		}
    		}
    		same = true;
    	}
    	int[] erg = new int[2];
    	erg[0] = errors;
    	erg[1] = abstentions;
    	return erg;
    }
	
    
    
    /**
	* Returns the number of wrong predictions by applying the given attributes and preferences to the data
	*@param data The data for evaluation
	*@param node The node containing the (grouped) attributes
	*@param preferences The preferences connected with the given attributes
	*@return The number of mistakes [0] induced by the given preference Rules and the number of abstentions [1]
	*/
	public int[] computeNodeErrors(Instances data, int[] tempAttributes, int[][] tempPreferences){
	boolean same = true;
		int errors = 0;
		int abstentions = 0;
		Instance one = new DenseInstance(data.numAttributes()/2);
		Instance two = new DenseInstance(data.numAttributes()/2);
		//pairwise comparison of the examples
		for(int i = 0; i < data.numInstances(); i++){
			for(int j =0; j < data.numAttributes()/2; j++){
				one.setValue(j, data.instance(i).value(j));
				two.setValue(j, data.instance(i).value(j+data.numAttributes()/2));
				if(one.value(j) != two.value(j)){
					same = false;
				}
			}
			//only compare outcomes that differ
			if(same != true){
				//Instance one is not preferred according to the given rule tree therefore wrong prediction
				if(decideNode(one, two, tempAttributes, tempPreferences) == 2){
					errors = errors +1;
					//abort since no improvement
					if(errorsGlobal > 0 && errorsGlobal < errors){
						int[] result = new int[2];
						result[0] = errors;
						result[1] = abstentions;
						return result;
      				}
				}
				else if (decideNode(one, two, tempAttributes, tempPreferences) == 0){
					abstentions = abstentions +1;
				}
			}
  		same = true;
		}
		int[] result = new int[2];
		result[0] = errors;
		result[1] = abstentions;
		return result;
	}
    
	
	
	
	/**
	 * Decides (internally) on the preference for the given instances
	 *@param one The first instance
	 *@param two The second instance to be compared with the first one
	 *@param stage The level of the attribute importance list based on which shall be decided
	 *@return 1 if instance one is preferred, 2 if two is preferred and 0 if equal/not comparable
	 */
	public int decide(Instance one, Instance two, int stage){
		//only consider the last stage of the given branch since the data is filtered in this way  
		//coordinates
		int i = stage;
		int jNode = 0;
		//first, compare the instances 
		boolean same = true; //Same values?
		for(jNode = 0; jNode < attributes[i].length; jNode++){
			if(one.value(attributes[i][jNode]) != two.value(attributes[i][jNode])){
				same = false;
	   		}
		}
		if(same == true){
			return 0;
	   	}
		//Missing values?
		for(jNode = 0; jNode < attributes[i].length; jNode++){
			if(one.isMissing(attributes[i][jNode])){
				return 0;
	   		}
			if(two.isMissing(attributes[i][jNode])){
				return 0;
	    	}
	   	}
		//check if instance one is preferred
		for(int rank = 0; rank < preferences[i].length; rank++){
			for(jNode = 0; jNode < attributes[i].length; jNode++){
				if(preferences[i][rank][jNode] != (int) one.value(attributes[i][jNode])){
					jNode = attributes[i].length;
	   	   		}
				else if(jNode == attributes[i].length-1){
					return 1;
	   	   		}
	   	   	}
			//check if instance two is preferred
			for(jNode = 0; jNode < attributes[i].length; jNode++){
				if(preferences[i][rank][jNode] != (int) two.value(attributes[i][jNode])){
					jNode = attributes[i].length;
	   	   		}
				else if(jNode == attributes[i].length-1){
					return 2;
	   	   		}
	   	   	}
		}
		return 0;
	}
	
	
	
	/**
	 * Decides (internally) on the preference for the given instances and node
	 *@param one The first instance
	 *@param two The second instance to be compared with the first one
	 *@param node The node containing the (grouped) attributes
	 *@param preferences The preferences connected with the given attributes
	 *@return 1 if instance one is preferred, 2 if two is preferred and 0 if equal/not comparable
	 */
	public int decideNode(Instance one, Instance two, int[] tempAttributes, int[][] tempPreferences){
		int jNode = 0;
		//compare the instances first
		boolean same = true; //Same values?
		for(jNode = 0; jNode < tempAttributes.length; jNode++){
			if(one.value(tempAttributes[jNode]) != two.value(tempAttributes[jNode])){
				same = false;
	   		}
		}
		if(same == true){
				return 0;
		}
		//Missing values?
		for(jNode = 0; jNode < tempAttributes.length; jNode++){
			if(one.isMissing(tempAttributes[jNode])){
   				return 0;
   			}
   			if(two.isMissing(tempAttributes[jNode])){
    			return 0;
    		}
   		}
   		//check if instance one is preferred
   		for(int rank = 0; rank < tempPreferences.length; rank++){
   			for(jNode = 0; jNode < tempAttributes.length; jNode++){
   	   			if(tempPreferences[rank][jNode] != (int) one.value(tempAttributes[jNode])){
   	   				jNode = tempAttributes.length;
   	   			}
   	   			else if(jNode == tempAttributes.length-1){
   	   				return 1;
   	   			}
   	   		}
   	   		//check if instance two is preferred
   	   		for(jNode = 0; jNode < tempAttributes.length; jNode++){
   	   			if(tempPreferences[rank][jNode] != (int) two.value(tempAttributes[jNode])){
   	   				jNode = tempAttributes.length;
   	   			}
   	   			else if(jNode == tempAttributes.length-1){
   	   				return 2;
   	   			}
   	   		}
		}
		return 0;
   }
	
	
	
	
	/**
     * Decides on the given instances based on the entire LP list
     *@param one The first instance
     *@param two The second instance to be compared with the first one
     *@param data The data for information on nominal or numeric attribute domain
     *@return 1 if instance one is preferred, 2 if two is preferred and 0 if equal/not comparable
     */
    public int decideExt(Instance one, Instance two, Instances data) throws Exception{
    	boolean same = true;
    	for(int j =0; j < one.numAttributes(); j++){
			if(one.value(j) != two.value(j)){
				same = false;
			}
		}
		//only compare outcomes that differ
		if(same != true){
			//start with the root node and go along the LP-List until first attribute X is reached s.t. o1[X] != o2[X]  
			for(int i = 0; i < attributes.length; i = i+1){
				//discretize values
				for(int j = 0; j < attributes[i].length; j++){
					if(cutPoints[i][j] != null && data.attribute(attributes[i][j]).isNumeric()){
						discretizeI(attributes[i][j], cutPoints[i][j], one);
        				discretizeI(attributes[i][j], cutPoints[i][j], two);
					}
				}
				int result = decideNode(one, two, attributes[i], preferences[i]);
				if(result != 0){
					if(result == 1){
						decidedAtNode[i][0]++;	
					}
					else{
						decidedAtNode[i][1]++;
					}
					return result;
				}
			}
			return 0;
		}
		else{
			return 0;
		}
    }
	
	
    
    
    /**
	* Checks whether the candidate attribute for labeling a node is already in branch 
	* (one attribute may only appear once per branch)
	* LP-structure: Conditional local preferences/conditional AI-tree
	* @param candidate The candidate attribute 
	* @param stage The stage in the tree
	* @return true if is included false if not
	*/
	public boolean inBranch(int candidate, int stage) {
		while(stage > 0){
	 		stage = stage-1;
	 		for(int jNode = 0; jNode < attributes[stage].length; jNode++){
	 			if(attributes[stage][jNode] == candidate){
	 	 			return true;
	 			}
	 		}
		}
		return false;
	}	
	
	
	
	
	/**
    * Compares the two given array values to the instance w.r.t. the given attributes
    *@param attributes The respective attributes
    *@param arrayOne The first array to be compared
    *@param arrayTwo The second array to be compared
    *@param instanceC The instance to be compared with the arrays
    *@return True if arrayOne and arrayTwo values are included false else
    */
    public static boolean compare(int[] attributes, int[] arrayOne, int[] arrayTwo, Instance instanceC){
    	//the cut value because of pairs of objects
   	  	int cut = instanceC.numAttributes()/2;
	   	//are the values of the arrays in line with the instance's values?
	   	for(int i = 0; i < attributes.length; i++){
	   		if(arrayOne[i] != (int) instanceC.value(attributes[i])){
	   			return false;
	     	}
	     	if(arrayTwo[i] != (int) instanceC.value(attributes[i] + cut) ){
	     		return false;
	     	} 
	   	}
	   	return true;
     }

	
    
    
     /**
     * Compares the values within the pairwise preference w.r.t. to the attributes given 
     *@param attributes The respective attributes to be compared
     *@param instanceC The instance to be compared w.r.t. to attributes
     *@return True if the values for the given attributes are the same False else
     */
     public static boolean sameVals(int[] attributes, Instance instanceC){
	   	  int cut = instanceC.numAttributes()/2;
	   	  //are the values of the arrays in line with the instance's values?
	   	  for(int i = 0; i < attributes.length; i++){
	   		  if(instanceC.value(attributes[i]) != instanceC.value(attributes[i] + cut)){
	   			  return false;
	   		  }
	   	  }
	   	  return true;
     }
    
    
     
     /**
      * Compares the given array values to the given instance
      *@param attributes The respective attributes
      *@param arrayOne The first array to be compared
      *@param instanceC The instance to be compared with the arrays
      *@return 1 if arrayOne values are included in the preferred instance -1 if included in the non-preferred instance and 0 else
      */
      public static int match(int[] attributes, int[] arrayOne, Instance instanceC){
    	  int cut = instanceC.numAttributes()/2;
    	  //are the values of the arrays in line with the instance's values?
    	  for(int i = 0; i < attributes.length; i++){
    		  if(arrayOne[i] != (int) instanceC.value(attributes[i])){
    			  i = attributes.length;
    		  }
    		  else if(i == attributes.length-1 ){
    			  return 1;
    		  }
    	  }
    	  for(int i = 0; i < attributes.length; i++){
    		  if(arrayOne[i] != (int) instanceC.value(attributes[i] + cut) ){
     			  i = attributes.length;
     		  } 
    		  else if(i == attributes.length-1){
    			  return -1;
    		  }
      		}
    	  return 0;
     }
     
      
      
      
    /**
  	* Resizes the attributes, cutPoints and preferences arrays to the given value
  	* @param size The new size for the arrays
  	*/
  	public void resize(int size) {
  		int[][] tmpAttributes = new int[size][];
  		int[][][] tmpPreferences = new int[size][][];
  		double[][][] tmpCutPoints = new double[size][][];
  		for(int i = 0; i < size; i ++){
  			//copy attributes
  			if(attributes[i] != null){
  				tmpAttributes[i] = new int[attributes[i].length];
  				for(int j = 0; j < attributes[i].length; j++){
  					tmpAttributes[i][j] = attributes[i][j];
  				}
  			}
  			//copy preferences
  			if(preferences[i] != null){
  				tmpPreferences[i] = new int[preferences[i].length][];
  				for(int j = 0; j < preferences[i].length; j++){
  					tmpPreferences[i][j] = new int[preferences[i][j].length];
  					for(int k = 0; k < preferences[i][j].length; k++){
  						tmpPreferences[i][j][k] = preferences[i][j][k];
  					}
  				}
  			}
  			//copy cuttinPoints
  			if(cutPoints[i] != null){
  				tmpCutPoints[i] = new double[cutPoints[i].length][];
  				for(int j = 0; j < cutPoints[i].length; j++){
  					tmpCutPoints[i][j] = new double[cutPoints[i][j].length];
  					for(int k = 0; k < cutPoints[i][j].length; k++){
  						tmpCutPoints[i][j][k] = cutPoints[i][j][k];
  					}
  				}
  			}
  		}
  		//resize attributes, preferences and cutpoints and copy tmp values
  		attributes = new int[size][];
  		preferences = new int[size][][];
  		cutPoints = new double[size][][];
  		for(int i = 0; i < size; i ++){
  			//copy attributes
  			if(tmpAttributes[i] != null){
  				attributes[i] = new int[tmpAttributes[i].length];
  				for(int j = 0; j < tmpAttributes[i].length; j++){
  					attributes[i][j] = tmpAttributes[i][j];
  				}
  			}
  			//copy preferences
  			if(tmpPreferences[i] != null){
  				preferences[i] = new int[tmpPreferences[i].length][];
  				for(int j = 0; j < tmpPreferences[i].length; j++){
  					preferences[i][j] = new int[tmpPreferences[i][j].length];
  					for(int k = 0; k < tmpPreferences[i][j].length; k++){
  						preferences[i][j][k] = tmpPreferences[i][j][k];
  					}
  				}
  			}
  			//copy cuttinPoints
  			if(tmpCutPoints[i] != null){
  				cutPoints[i] = new double[tmpCutPoints[i].length][];
  				for(int j = 0; j < tmpCutPoints[i].length; j++){
  					cutPoints[i][j] = new double[tmpCutPoints[i][j].length];
  					for(int k = 0; k < tmpCutPoints[i][j].length; k++){
  						cutPoints[i][j][k] = tmpCutPoints[i][j][k];
  					}
  				}
  			}
  		}
  	}	    
  	  
      
     
         
    /**
     * Returns the set of all possible combinations for the given number of attributes and max grouping size
     *@param number The maximal "grouping size" (=maximal amount of attributes per set)
     *@param index The index of the last attribute to be considered (assumption: A1, A2,... AIndex)
     *@return Array containing the set of all combinations
     */
    public static int[][] combinations(int number, int index){
    	//determine the necessary size of the array (= number of possible attribute combinations)
   	  	int size = 0;
   	  	for(int i = 0; i < number; i++){
   	  		size = size + cardinality (i+1, index);
   	  	}
  	  	int[][] set = new int[size][];
   	  	//navigating first coordinate of order (the rank)
   	  	int i;
   	  	int tmp = index;
   	  	for(i = 0; i < tmp; i ++){
    		  set[i] = new int[1];
    		  set[i][0] = i+1;
        }
        int count = 0;
    	//saving the different attribute combinations for more dimensions (second dimension = attribute combination)
      	for(int k = 1 ; k < number; k++){
      	  //initialize all values for set of i attributes per combination
      	  tmp = cardinality (k+1, index);
      	  //compute all values for k+1 dim
          for(int l = 0; l < tmp; l ++){
        	  int start = set[count][k-1];
        	  if(start == index){
        		  l = l-1;
        	  }
        	  //"add" all possible attribute combinations to the first k-th attribute combinations adding count's combination 
        	  for(int j = start; j < index; j++){
        		  set[i] = new int[k+1];
        		  //initialize the start value(s) from the 
        		  for(int m = 0; m < k; m++){
        			  set[i][m] = set[count][m];
        		  }
        		  if(j > start){
        			set[i][k] = set[i-1][k]+1;
        			l++;
        		  }
        		  else{
        			set[i][k] = set[count][k-1]+1;
        		  }
        		  i++;
        	  }
          	  count ++;
          }
          //determine the correct position for count
          count = 0;
          for (int m = 0; m < k; m++){
          	  count = count + cardinality (m+1, index);
          }
      	}
      	return set;
    }
    
    
    
    
    /**
      * Returns the cardinality for the given set (disorderly drawing without lay back)
      *@param number The amount of "drawings"
      *@param index The number of objects to be considered
      *@return Array containing the set of all combinations
      */
    public static int cardinality(int number, int index){
      //for number == index the result is defined as one (prevents division by zero)
      if(number == index){
    	  return 1;
      }
     	  
      int size = 1;
      for(int i = 1; i <= number; i++){
     	  size = size*(index-number+i)/i;
      }
      return size;
    }
       
       
       
       
     /**
      * Returns the distinct values (double) for the given attribute in the set of pairwise preferences
      *@param data The data for exploration (set of pairwise preferences)
      *@param attr The attributes for exploration
      *@return Array filled with the labels for given attributes ([attr][labels of attr])
     */
     public static double[] distAttrDoubleVals(Instances data, int attr){
	   	int cut = data.numAttributes()/2;
	   	Instance temp = new DenseInstance(data.numAttributes());
	   	//create the training data s.t. attribute values of second outcome will be considered, too
	   	Instances training = new Instances(data, data.numInstances()*2);
	   	for(int i = 0; i < data.numInstances(); i++){
	   		training.add(data.instance(i));
	   		temp = (Instance) data.instance(i).copy();
	   		temp.setValue(attr, temp.value(attr + cut));
	   		temp.setValue(attr + cut, data.instance(i).value(attr));
	   		training.add(temp);
	   	}
	   	//determine for each attribute distinct values
	   	double[] distVals;
	   	distVals = new double[training.numDistinctValues(attr)];
	   	if (training.attribute(attr).isNumeric()) {
	        double[] attVals = training.attributeToDoubleArray(attr);
	        Arrays.sort(attVals);
	        //Identify distinct attribute values
	        int counter = 1;
	        distVals[counter-1] = attVals[counter-1];
	        for (int i = 1; i < attVals.length; i++) {
	        	if(distVals[counter-1] != attVals[i]){
	        		distVals[counter] = attVals[i];
	         		counter++;
	          	}
	        }
	   	}
	    else{
        	distVals = new double[data.attribute(attr).numValues()];
        	Enumeration enu= data.attribute(attr).enumerateValues();
      		int k = 0;
      		String tempor; 
      		double transl; 
      		while(enu.hasMoreElements()) {
      			tempor = (String) enu.nextElement();
      			transl = Double.parseDouble(tempor) ;
      			distVals[k] = transl;
      			k++;
      		}
	    }
	    return distVals;
    }
	       
	       
     
        
    /**
      * Returns the attributes' ordered labels based on the pairwise preferences (according to Copeland's method)
      *@param training The data for exploration
      *@param attribute The given attributes
      *@return Array filled with the attribute's ordered labels ([rank][label of attr 1, label of attr 2,...])
      */
    public int[][] orderedAttrVals(Instances training, int attribute[]){
      	//Distinct values of the given attribute
    	int[][] distVals = distAttrVals(training, attribute);
        int vals = 1;
        int tmp;
        //determine whether there is a monotone attribute to be considered
        boolean mon = false;
        int r = 0;
        for(int i = 0; i < attribute.length; i++){
        	if(monotony[attribute[i]]>0){
        		mon = true;
        		r++;
        	}
        }
    	//at least monotony for one attribute, therefore monotony needs to be respected (i.e. non-compliant orders will be removed afterwards
        if(mon == true){
        	//compute the number of possible value combinations
        	for(int i = 0; i < distVals.length; i ++){
        		vals = vals * distVals[i].length;
        	}
        	// variables for determining the order labels
        	int[][] order = new int[vals][attribute.length];
        	int[] position = new int[attribute.length];
        	//initialize all values for first rank
        	for(int j = 0; j < order[0].length; j ++){
        		order[0][j] = distVals[j][0];
        		position[j] = 0;
        	}
        	int i = 1;
        	//navigating the second coordinate of order (the attribute (value) corresponding to a rank)
        	int j = order[0].length - 1;
        	//navigating the second coordinate of distVals (the value of the corresponding attribute)
        	int l = 1;
        	//computing and saving all possible attribute values until all attributes are considered
        	while(j >= 0){
        		while(j == distVals.length -1 && l < distVals[j].length ){
        			//initialize the array by the array values before
        			System.arraycopy(order[i-1], 0, order[i], 0, order[i-1].length);
        			order[i][j] = distVals[j][l];
        			l = l+1;
        			i = i+1;
        		}
        		l = 1;
        		j = j-1;
        		if(i < vals){
        			if(position[j] < distVals[j].length-1){
        				for(int k = j+1; k < attribute.length; k++ ){
        					position[k] = 0;
        				}
        				position[j] = position[j]+1;
        				//set the values to start before attribute j and leave values after 
        				for(int k = 0; k < attribute.length; k++){
        					order[i][k] = distVals[k][position[k]];
        	       		}
        				i = i+1;
        				j = distVals.length-1;
        			}
        			//all values for attribute j considered, next attribute j-1 
        		}
        	}
        	//Matrix in which the entries represent the occurrence of a certain (pairwise) preference
        	int[][] frequency = new int[vals][vals];
        	//filling the matrix with 0 as default value
        	for(i = 0;i < vals; i++){
        		for(j = 0 ; j < vals;j++){
        			frequency[i][j] = 0;
        		}
        	}
        	// pairwise comparison of outcome one to outcome two within an instances 
        	for(i = 0; i < training.numInstances(); i++){
        		for(int k = 0; k < vals; k++){
        			for(l = k+1; l < vals; l++){
        				if(compare(attribute, order[k], order[l], training.instance(i)) ){
        					frequency[k][l]++;
        				}
        				else if(compare(attribute, order[l], order[k], training.instance(i)) ){
        					frequency[k][l]--;
        				}
        			}
        		}
        	}
        	int[] frequencies = new int[vals];
        	for(int lbl = 0; lbl <vals; lbl++){
        		tmp = 0;
        		for(i = 1; i < vals; i++){
        			if( lbl < vals && (frequency[lbl][i] ) > 0){
        				tmp = tmp + frequency[lbl][i];
        	   	 	}
        			if( (frequency[i-1][lbl]) < 0){
        				tmp = tmp - frequency[i-1][lbl];
        			}
        		}
        		frequencies[lbl] = tmp;
        	}
        	//sort the attribute labels in accordance with the aggregated victories/defeats
        	int cache;
        	int[] cacheA = new int[order[0].length];
        	for(i = 0; i < vals-1; i++){
        		if(frequencies[i] < frequencies[i+1]){
        			cache = frequencies[i];
        			frequencies[i] = frequencies[i+1];
        			frequencies[i+1] = cache;
        			System.arraycopy(order[i], 0, cacheA, 0, order[0].length);
        			System.arraycopy(order[i+1], 0, order[i] , 0, order[0].length);
        			System.arraycopy(cacheA, 0, order[i+1] , 0, order[0].length);
        			i = -1;
        		}
        	}
        	//respect given monotony on attribute values
        	for(int k = 0; k < attribute.length; k++){
        		if(monotony[attribute[k]] == 1){
        			for(int h = 0; h < order.length-1; h++){
        				//swap values if not decreasing
        				if(order[h][k] < order[h+1][k]){
        					//swap for complete combo
        					for(int m = 0; m < attribute.length; m++){
        						int tmpVal = order[h][m];
        						order[h][m] =order[h+1][m];
        						order[h+1][m] = tmpVal;
        					}
        					h=-1;
        				}
        			}
        		}
        		if(monotony[attribute[k]] == 2){
        			for(int h = 0; h < order.length-1; h++){
        				//swap values if not increasing
        				if(order[h][k] > order[h+1][k]){
        					//swap for complete combo
        					for(int m = 0; m < attribute.length; m++){
        						int tmpVal = order[h][m];
        						order[h][m] = order[h+1][m];
        						order[h+1][m] = tmpVal; 
        					}
        					h=-1;
        				}
        			}  
        		}
        	}
        	return order;
        }
        //no monotone attributes need to be considered
        else{
        	for(int i = 0; i < distVals.length; i ++){
        		vals = vals * distVals[i].length;
        	}
        	int[][] order = new int[vals][attribute.length];
        	//saving the different values within the array (first dimension = rank; second dimension = attribute values' combination)
        	int[] position = new int[attribute.length];
        	for(int j = 0; j < order[0].length; j ++){
        		order[0][j] = distVals[j][0];
        		position[j] = 0;
        	}
        	int i = 1;
        	int j = order[0].length - 1;
        	int l = 1;
        	//computing and saving all possible attribute values until all attributes are considered
        	while(j >= 0){
        		while(j == distVals.length -1 && l < distVals[j].length ){
        			System.arraycopy(order[i-1], 0, order[i], 0, order[i-1].length);
        			order[i][j] = distVals[j][l];
        			l = l+1;
        			i = i+1;
        		}
        		l = 1;
        		j = j-1;
        		if(i < vals){
        			if(position[j] < distVals[j].length-1){
        				for(int k = j+1; k < attribute.length; k++ ){
        					position[k] = 0;
        				}
        				position[j] = position[j]+1;
        				//set the values to start varlues before attribute j and leave values after 
        				for(int k = 0; k < attribute.length; k++){
        					order[i][k] = distVals[k][position[k]];
        	       		}
        				i = i+1;
        				j = distVals.length-1;
        			}
        		}
        	}
        	//Matrix in which the entries represent the occurrence of a certain (pairwise) preference
      	   	int[][] frequency = new int[vals][vals];
      	   	//filling the matrix with as default value
      	   	for(i = 0;i < vals; i++){
      	   		 for(j = 0 ; j < vals;j++){
      	   			 frequency[i][j] = 0;
      	   		 }
      	   	}
      	   	 for(i = 0; i < training.numInstances(); i++){
      			 for(int k = 0; k < vals; k++){
      				 for(l = k+1; l < vals; l++){
	      	   			 if( compare(attribute, order[k], order[l], training.instance(i)) ){
	      	   				 frequency[k][l]++;
	      	   			 }
	      	   			 else if(compare(attribute, order[l], order[k], training.instance(i)) ){
	      	   				 frequency[k][l]--;
	      	   			 }
      				 }
      			}
      	   	}
      	   	//compute the label-order for output
      	   	int[] frequencies = new int[vals];
      	   	//sum up for every label victories/defeats
      	   	for(int lbl = 0; lbl <vals; lbl++){
      	   		tmp = 0;
      	   	 	for(i = 1; i < vals; i++){
      	   	 		if( lbl < vals && (frequency[lbl][i] ) > 0){
      	   	 			tmp = tmp + frequency[lbl][i];
      	   	 		}
      	   	 		if( (frequency[i-1][lbl]) < 0){
      	   	 				tmp = tmp - frequency[i-1][lbl];
      	   	 			}
      	   	 		}
      	   	 		frequencies[lbl] = tmp;
      	   	 	}
      	     //return the order w.r.t. victories/defeats
      	   	 int cache;
      	   	 int[] cacheA = new int[order[0].length];
      	      	 for(i = 0; i < vals-1; i++){
      	      		 if(frequencies[i] < frequencies[i+1]){
      	      			 cache = frequencies[i];
      	      			 frequencies[i] = frequencies[i+1];
      	      			 frequencies[i+1] = cache;
      	      			 System.arraycopy(order[i], 0, cacheA, 0, order[0].length);
      	      			 System.arraycopy(order[i+1], 0, order[i] , 0, order[0].length);
      	      			 System.arraycopy(cacheA, 0, order[i+1] , 0, order[0].length);
      	      			 i = -1;
      	      		 }
      	      	 }
   	        return order;
     	}
     }	
    
    
    
    
    /**
     * Returns the distinct values for the given attributes in the set of pairwise preferences
     *@param data The data for exploration (set of pairwise preferences)
     *@param attr The attributes for exploration
     *@return Array filled with the labels for given attributes ([attr][labels of attr])
     */
     public static int[][] distAttrVals(Instances data, int[] attr){
	    	int cut = data.numAttributes()/2;
	    	Instance temp = new DenseInstance(data.numAttributes());
	     	//create the training data s.t. attribute values of second outcome will be considered, too
	    	Instances training = new Instances(data, data.numInstances()*2);
	    	for(int i = 0; i < data.numInstances(); i++){
	    		training.add(data.instance(i));
	    		//switch attribute attr and attr+cut values to consider the second outcome
	    		temp = (Instance) data.instance(i).copy();
	    		for(int j = 0; j < attr.length; j++){
	    			temp.setValue(attr[j], temp.value(attr[j]+cut));
	    			temp.setValue(attr[j]+cut, data.instance(i).value(attr[j]));
	    		}
	    		training.add(temp);
	    	}
	    	//determine for each attribute distinct values
	    	int[][] distVals = new int[attr.length][];
	    	for(int j = 0; j < attr.length; j++){
	    		distVals[j] = new int[training.numDistinctValues(attr[j])];
	    		if (training.attribute(attr[j]).isNumeric()) {
	       	      	double[] attVals = training.attributeToDoubleArray(attr[j]);
	       	      	int [] sorted = Utils.sort(attVals);
	       	      	double prev = 0;
	       	      	int counter = 0;
	       	      	//Identify distinct attribute values
	       	      	for (int i = 0; i < sorted.length; i++) {
	       	      		Instance current = training.instance(sorted[i]);
	       	      		if (current.isMissing(attr[j])) {
	       	      			break;
	       	      		}
	       	      		if ((i == 0) || (current.value(attr[j]) > prev)) {
	       	      			prev = current.value(attr[j]);
	       	      			distVals[j][counter] = (int) current.value(attr[j]);
	       	      			counter++;
	       	      		} 
	       	      	}
	       	      	if(training.numDistinctValues(attr[j]) == 0){
	       	      		//no value occurred for this attribute 
	       	      		distVals[j] = new int[1];
	       	      		//99 stands for not occurred
	       	      		distVals[j][0] = 99;
	       	      	}
	    		} 
	    		else 
	    		{
	    			distVals[j] = new int[training.attribute(attr[j]).numValues()];
	    			Enumeration enu= training.attribute(attr[j]).enumerateValues();
	    			int k = 0;
	    			String tempor; 
	    			int transl; 
	    			while(enu.hasMoreElements()) {
	    				tempor = (String) enu.nextElement();
	    				transl = Integer.parseInt(tempor);
	    				distVals[j][k] = transl-1;
	    				k++;
	    			}
	    		}
	    	}
	    	return distVals;
     }
     
     
     
     
     /**
      * Discretizes the values of attribute attr according to the given cutting points
      * @param attr The attribute for which the values shall be discretized
      * @param cutPoints The cutting points for the discretization
      * @param data The data to be discretized (for attribute attr) (set of pairwise instances!)
      * @return The data set with discretized attribute values
      */
     public static Instances discretize(int attr, double[] cutPoints, Instances data)throws Exception{
     	int mid = data.numAttributes()/2;
     	//discretize the attributes' values in the set of pairwise instances according to the cutting points
     	for(int i = 0; i < data.numInstances(); i++){
     		for(int j = 0; j < cutPoints.length-1; j++){
     			// for attribute value between [cuttingPoint[j], cuttingPoint[j+1] set the value to (discrete) value (j+1)
     			if( (data.instance(i).value(attr) >= cutPoints[j]) && ( data.instance(i).value(attr) <= cutPoints[j+1]) ){
     				data.instance(i).setValue(attr, (j+1));
     				j = cutPoints.length;
     			}
     		}
     		//same procedure for the second attribute value of the pairwise instance
     		for(int j = 0; j < cutPoints.length-1; j++){
     			// for attribute value between [cuttingPoint[j], cuttingPoint[j+1] set the value to (discrete) value (j+1)
     			if( (data.instance(i).value(attr + mid) >= cutPoints[j]) && ( data.instance(i).value(attr + mid) <= cutPoints[j+1]) ){
     				data.instance(i).setValue( (attr+mid), (j+1));
     				j = cutPoints.length;
     			}
     		}
     	}
     	return data;
     }
     
     
     
     
     /**
      * Discretize the value of attribute attr according to the given cutting points within an instance
      * @param attr The attribute for which the values shall be discretized
      * @param cutPoints The cutting points for the discretization
      * @param inst The instance to be discretized (for attribute attr) 
      * @return The instance with discretized attribute value
      */
     public static Instance discretizeI(int attr, double[] cutPoints, Instance inst)throws Exception{
     	//discretize the attributes' value within the instance to the cutting points
    	for(int j = 0; j < cutPoints.length-1; j++){
    		// for attribute value between [cuttingPoint[j], cuttingPoint[j+1] set the value to (discrete) value (j+1)
    		if( (inst.value(attr) >= cutPoints[j]) && ( inst.value(attr) <= cutPoints[j+1]) ){
    			inst.setValue(attr, (j+1));
    			//exit loop
    			j = cutPoints.length;
    		}
    	}
     	return inst;
     }
        	 
     
     
     
     /**
      * Computes the cutting points for the given data set according to Fayyad & Irani extended to a set of pairwise preferences/instances 
      * @param attr The attribute for consideration 
      * @param data The set of pairwise instances
      * @param boundary Shall only boundary cutting be considered? 
      * @param stop The maximum number of bins 
      * @return The cutting points required for the following discretization of values for the given attribute 
      */
     public static double[] cutPoints(int attr, Instances data, boolean boundary, int stop){
     	Instances tmp = new Instances(data);
     	int mid = tmp.numAttributes()/2;
     	//create the set for computing the cutting points (switch first and second attribute value of attr if second is greater, create respective class indices...)
     	//eliminate attributes not needed
     	for(int i = 0; i < mid; i++){
     		if(i != attr){
     			tmp.deleteAttributeAt(mid+i);
     			tmp.deleteAttributeAt(i);
     			mid = mid - 1;
     			i = i -1;
     			attr = attr -1;
     		}
     	}
     	//create the class attribute and class value
     	tmp.insertAttributeAt(new Attribute("classAttr"), 2);
     	for(int i = 0; i < tmp.numInstances(); i++){
     		//switch attribute values if necessary (considering symmetry for evaluation)
     		if(tmp.instance(i).value(0) < tmp.instance(i).value(1)){
     			double tmpVal = tmp.instance(i).value(0);
     			tmp.instance(i).setValue(0, tmp.instance(i).value(1));
     			tmp.instance(i).setValue(1, tmpVal);
     			tmp.instance(i).setValue(2, 0);
     		}
     		else{
     			tmp.instance(i).setValue(2,1);
     		}
     	}
     	//initiate the set of cutting points
     	double[] cutPoint;
     	cutPoint = new double[2];
     	cutPoint[0] = 0;
     	//determine the maximum attribute value in the data set
     	tmp.sort(0);
     	double max = tmp.instance(tmp.numInstances()-1).value(0);
     	tmp.sort(1);
     	if(tmp.instance(tmp.numInstances()-1).value(1) > max){
     		max = tmp.instance(tmp.numInstances()-1).value(1);
     	}
     	cutPoint[1] = max;
     	//identify potential (boundary?) cutting points 
     	Instances potCutPoints = potCutPoints(tmp, boundary);
     	//if there are no boundary cutting points consider non-boundary cutting points
     	if(potCutPoints.numInstances() == 0){
     		potCutPoints = potCutPoints(tmp, false);
     	}
     	double optCut;
     	//identify the optimal cutting points in the set of potential points (number of cutting points is limited to stop)
     	for(int i = 0; i < stop; i++){
     		optCut = optCuttingPoint(tmp, potCutPoints, cutPoint);
     		if(optCut == -1){
     			return cutPoint;
     		}
     		else{
     			//add the new cutting point and sort the array
     			double[] temp = new double[cutPoint.length];
             	System.arraycopy( cutPoint, 0, temp, 0, cutPoint.length );
         		cutPoint = new double[cutPoint.length+1];
         		System.arraycopy( temp, 0, cutPoint, 0, temp.length );
             	cutPoint[cutPoint.length-1] = optCut;
         		Arrays.sort(cutPoint);
     		}
     	}
     	return  cutPoint;
     }
     
     
    
     /**
      * Returns the possible (boundary) cutting points for the given data set
      * @param inst The pairwise set of attribute value to be evaluated
      * @param boundary Returns only boundary cutting points if true
      * @return The set resulting cutting points
      */
     public static Instances potCutPoints(Instances inst, boolean boundary){
    	//Initialize the set
     	Attribute Attribute1 = new Attribute("Cut_Point");
     	ArrayList<Attribute> fvWekaAttributes = new ArrayList<Attribute>(1);
     	fvWekaAttributes.add(Attribute1);
     	Instances points = new Instances("Cut_Points_Rela", fvWekaAttributes, 0);
     	inst.sort(1);
     	inst.sort(0);
      	for (int i = 0; i < inst.numInstances()-1;i++){
     		if(boundary == true){
     			if( inst.instance(i).value(0) != inst.instance(i+1).value(0) && inst.instance(i).value(2) != inst.instance(i+1).value(2)){
     				double mid = inst.instance(i).value(0)- inst.instance(i+1).value(0);
     				mid = mid /2 + inst.instance(i+1).value(0);
     				Instance midpoint = new DenseInstance(1);
     				midpoint.setValue((Attribute)fvWekaAttributes.get(0), mid);
     				if(points.numInstances() > 0){
     					double diff = midpoint.value(0)-points.lastInstance().value(0);
     					diff = Math.abs(diff);
         				if(diff > 0.1){
         					points.add(midpoint);
         				}
     				}else{
     					points.add(midpoint);
     				}
     			}
     		}
     		else{
     			if( inst.instance(i).value(0) != inst.instance(i+1).value(0)){
     				double mid = inst.instance(i).value(0)- inst.instance(i+1).value(0);
     				mid = mid /2 + inst.instance(i+1).value(0);
     				Instance midpoint = new DenseInstance(1);
     				midpoint.setValue((Attribute)fvWekaAttributes.get(0), mid);
     				points.add(midpoint);
     			}
     		}
     	}
     	//consider second coordinate's attribute values for cutting points
     	inst.sort(1);
     	for (int i = 0; i < inst.numInstances()-1;i++){
     		if(boundary == true){
     			if( inst.instance(i).value(1) != inst.instance(i+1).value(1) && inst.instance(i).value(2) != inst.instance(i+1).value(2)){
     				double mid = inst.instance(i).value(1)- inst.instance(i+1).value(1);
     				mid = mid /2 + inst.instance(i+1).value(1);
     				Instance midpoint = new DenseInstance(1);
     				midpoint.setValue((Attribute)fvWekaAttributes.get(0), mid);
     				points.add(midpoint);
     			}
     		}
     		else{
     			if( inst.instance(i).value(1) != inst.instance(i+1).value(1)){
     				double mid = inst.instance(i).value(1)- inst.instance(i+1).value(1);
     				mid = mid /2 + inst.instance(i+1).value(1);
     				Instance midpoint = new DenseInstance(1);
     				midpoint.setValue((Attribute)fvWekaAttributes.get(0), mid);
     				points.add(midpoint);
     			}
     		}
     	}
     	return points;
     }
     
     
     
     
     /**
      * Evaluate the set of possible cutting Points given and return the best one in terms of entropy (minimal) given the set of cut points
      * @param inst The pairwise set of attribute values to be evaluated
      * @param points The set of possible cutting points
      * @param cutPoints The set of cutting points already identified
      * @return The optimal cutting point to be added; output -1 indicates that there was no improvement in terms of class information entropy 
      */
     public static double optCuttingPoint(Instances inst, Instances points, double[] cutPoints){
     	double optimum;
     	double[] newCutPoints = new double[cutPoints.length+1];
     	//class information entropy of the partition
     	double partitionEntropy[] = new double[points.numInstances()];
     	//compute for every candidate cutting point the resulting average entropy and choose that candidate that minimizes the entropy
     	for(int i = 0; i < points.numInstances(); i++){
         	System.arraycopy( cutPoints, 0, newCutPoints, 0, cutPoints.length );
     		newCutPoints[newCutPoints.length-1] = points.instance(i).value(0);
     		Arrays.sort(newCutPoints);
     		//class information entropy of partition / average the entropy
     		partitionEntropy[i] = classInfEntropy(newCutPoints, inst);
     	}
     	//determine the candidate cutting point with minimal entropy
     	double min = partitionEntropy[0];
     	int pos = 0;
     	for(int i = 1; i < partitionEntropy.length; i++){
     		if(partitionEntropy[i] < min){
     			min = partitionEntropy[i];
     			pos = i;
     		}
     	}
     	optimum = points.instance(pos).value(0);
     	//add the new cutting point and sort the array
     	System.arraycopy( cutPoints, 0, newCutPoints, 0, cutPoints.length );
 		newCutPoints[newCutPoints.length-1] = optimum;
 		Arrays.sort(newCutPoints);
     	//compute the class information entropy of the partition before new split and after to decide whether to refrain or not the new cutting point
     	//entropy before (optimal) split
     	double entropyBefore = classInfEntropy(cutPoints, inst);
     	double entropyAfter = classInfEntropy(newCutPoints, inst);
     	if(entropyBefore <= entropyAfter){
     		return -1;
     	}
     	double gain =  entropyBefore - entropyAfter;	
     	double[] singleEntropyBefore = entropy(cutPoints, inst);
     	double[] singleEntropyAfter = entropy(newCutPoints, inst);
     	double resEntropyBefore = 0;
     	double resEntropyAfter = 0;
     	for(int i = 0; i < singleEntropyBefore.length; i++){
     		resEntropyBefore = resEntropyBefore + singleEntropyBefore[i];
     	}
     	for(int i = 0; i < singleEntropyAfter.length; i++){
     		resEntropyAfter = resEntropyAfter + singleEntropyAfter[i];
     	}
     	double threshold;
     	double deltaATS;
     	deltaATS = lb(7) - 2*(resEntropyBefore - resEntropyAfter);
     	threshold = (lb(inst.numInstances()-1) + deltaATS)/inst.numInstances();
     	if(gain <= threshold){
     		return -1;
     	}
     	return optimum;
     }
     
     
     
     
     /**
      * Compute the class information entropy for the resulting partition based on the cutting points
      * @param cutPoints The cutting points
      * @param inst The instances
      * @return The class entropy resulting from cutting according to the given points for the given set of instances
      */
     public static double classInfEntropy(double[] cutPoints, Instances inst){
    	double classInfEntropy;
     	//the class distribution given the identified cutting points within the resulting intervals (needed for computing class entropy)
     	int [][] classDistribution;
     	//the amount of sets is +1 because one needs to consider the new cutting to be added, too
     	int amount = (cutPoints.length * (cutPoints.length+1))/2;
     	//the class distribution over the sets resulting from the cutting points #class 0 in set = [set][0] and #class 1 in set = [set][1]
     	classDistribution = new int[amount][2];
     	//compute the resulting interval combinations/borders
 		double[][] intervals = intervalComb(cutPoints);
 		double[] entropy = entropy(cutPoints, inst);
 		//determine the class distributions 
 		for(int j = 0; j < inst.numInstances(); j++){
 			//check whether the instance j lies in the interval x interval
 			for(int k = 0; k < intervals.length; k++){
 				if( ((inst.instance(j).value(0) >= intervals[k][0]) && (inst.instance(j).value(0) <= intervals[k][1])) && ((inst.instance(j).value(1) >= intervals[k][2]) && (inst.instance(j).value(1) <= intervals[k][3])) ){
 					// instance / point is within the set, store the class value for the distribution
 					if(inst.instance(j).value(2) == 0){
 						classDistribution[k][0] ++;
 					}
 					else{
 						classDistribution[k][1] ++;
 					}
 				}
 			}
 		}
  		//class information entropy of partition / average the entropy
 		double avEntropy = 0;
 		for(int k = 0; k < entropy.length; k++){
 			avEntropy = avEntropy + entropy[k] * ( (double)(classDistribution[k][0]+ classDistribution[k][1])/ (double) inst.numInstances());
 		}
 		classInfEntropy = avEntropy;
     	return classInfEntropy;
     }
     
     
     
     
     /**
      * Compute the class entropy given the cutting points for every resulting set
      * @param cutPoints The cutting points
      * @param inst The instances
      * @return The class entropy resulting from cutting according to the given points for every set
      */
     public static double[] entropy(double[] cutPoints, Instances inst){
     	//the class distribution given the identified cutting points within the resulting intervals (needed for computing class entropy)
     	int [][] classDistribution;
     	//the amount of sets resulting from the cutting points
     	int amount = (cutPoints.length * (cutPoints.length-1))/2;
     	//the class distribution over the sets resulting from the cutting points #class 0 in set = [set][0] and #class 1 in set = [set][1]
     	classDistribution = new int[amount][2];
     	double entropy[] = new double[amount];
     	//compute the resulting interval combinations/borders
 		double[][] intervals = intervalComb(cutPoints);
 		//determine the class distributions 
 		for(int j = 0; j < inst.numInstances(); j++){
 			//check whether the instance j is included in interval x
 			for(int k = 0; k < intervals.length; k++){
 				if( ((inst.instance(j).value(0) >= intervals[k][0]) && (inst.instance(j).value(0) <= intervals[k][1])) && ((inst.instance(j).value(1) >= intervals[k][2]) && (inst.instance(j).value(1) <= intervals[k][3])) ){
 					// instance/point is within the set, store the class value for the distribution
 					if(inst.instance(j).value(2) == 0){
 						classDistribution[k][0] ++;
 					}
 					else{
 						classDistribution[k][1] ++;
 					}
 				}
 			}
 		}
     	//compute the class entropy for the sets
 		for(int k = 0; k < intervals.length; k++){
 			double proportionZero = 0;
 			if(classDistribution[k][0] > 0){
 				proportionZero = (double) classDistribution[k][0] / (classDistribution[k][0]+ classDistribution[k][1]); 
     		}
 			double proportionOne = 0;
 			if(classDistribution[k][1] > 0){
 				proportionOne  = (double) classDistribution[k][1] / (classDistribution[k][0]+ classDistribution[k][1]);
 			}
 			if(proportionZero > 0){
 				entropy[k] = - (proportionZero * lb(proportionZero));
 			}
 			else{
 				entropy[k] = 0;
 			}
 			if(proportionOne > 0){
 				entropy[k] = entropy[k] - (proportionOne * lb(proportionOne));
 			}
 		}
     	return entropy;
     }
     
     
     
     
     /**
      * Compute every possible (in this context) combination of two intervals generated by a given set of points
      * @param points The points for the interval borders
      * @return The possible combinations of intervals
      */
     public static double[][] intervalComb(double[] points){
     	//compute the number of possible combinations
     	int amount = (points.length * (points.length-1))/2;
     	//store the possible combinations [amount combos][coordinates]
     	double[][] combos = new double [amount][4];
     	int count = 0;
     	for(int i = 0; i < points.length-1; i++){
     		for(int j = 0; j <= i; j++){
     			//first interval
         		combos[count][0] = points[i];
         		combos[count][1] = points[i+1];
         		//second inverval
     			combos[count][2] = points[j];
     			combos[count][3] = points[j+1];
     			count++;
     		}
     	}
     	return combos;
     }	
     
     
     
     
     /**
      * Returns the logarithm with base 2
      * @param x The argument for the computation
      * @return lb(x)
      */
     public static double lb(double x){
         return Math.log(x) / Math.log(2.0);
     }
     
    
     
}
